"""Constants for the onboarding component."""
DOMAIN = "onboarding"
STEP_USER = "user"
STEP_CORE_CONFIG = "core_config"
STEP_INTEGRATION = "integration"
STEP_ANALYTICS = "analytics"
STEP_MOB_INTEGRATION = "mob_integration"

STEPS = [
    STEP_USER,
    STEP_CORE_CONFIG,
    STEP_ANALYTICS,
    STEP_INTEGRATION,
    STEP_MOB_INTEGRATION,
]

DEFAULT_AREAS = ("living_room", "kitchen", "bedroom")
