"""Constants for the Bond integration."""

BRIDGE_MAKE = "Olibra"

DOMAIN = "bond"

CONF_BOND_ID: str = "bond_id"


HUB = "hub"
BPUP_SUBS = "bpup_subs"

SERVICE_SET_FAN_SPEED_TRACKED_STATE = "set_fan_speed_tracked_state"
SERVICE_SET_POWER_TRACKED_STATE = "set_switch_power_tracked_state"
SERVICE_SET_LIGHT_POWER_TRACKED_STATE = "set_light_power_tracked_state"
SERVICE_SET_LIGHT_BRIGHTNESS_TRACKED_STATE = "set_light_brightness_tracked_state"
ATTR_POWER_STATE = "power_state"
