#!/bin/sh
echo "echo -e ''" > /data/data/pl.sviete.dom/files/usr/bin/sudo 
echo "echo -e '\e[38;5;220m Cześć tu \e[30;48;5;208m Jolka \e[0m'" >> /data/data/pl.sviete.dom/files/usr/bin/sudo 
echo "echo -e '\e[38;5;220m komenda sudo została zablokowana dla Twojego dobra\e[0m'" >> /data/data/pl.sviete.dom/files/usr/bin/sudo 
echo "echo -e ''" >> /data/data/pl.sviete.dom/files/usr/bin/sudo 
echo "2021-12-28" > /data/data/pl.sviete.dom/files/home/AIS/.ais_fix
