"""Constants for VeSync Component."""

DOMAIN = "vesync"
VS_DISPATCHERS = "vesync_dispatchers"
VS_DISCOVERY = "vesync_discovery_{}"
SERVICE_UPDATE_DEVS = "update_devices"

VS_SWITCHES = "switches"
VS_FANS = "fans"
VS_LIGHTS = "lights"
VS_MANAGER = "manager"
