"""Constants for the Mill heater component."""

ATTR_AWAY_TEMP = "away_temp"
ATTR_COMFORT_TEMP = "comfort_temp"
ATTR_ROOM_NAME = "room_name"
ATTR_SLEEP_TEMP = "sleep_temp"
BATTERY = "battery"
CLOUD = "Cloud"
CONNECTION_TYPE = "connection_type"
CONSUMPTION_TODAY = "day_consumption"
CONSUMPTION_YEAR = "year_consumption"
DOMAIN = "mill"
ECO2 = "eco2"
HUMIDITY = "humidity"
LOCAL = "Local"
MANUFACTURER = "Mill"
MAX_TEMP = 35
MIN_TEMP = 5
SERVICE_SET_ROOM_TEMP = "set_room_temperature"
TEMPERATURE = "current_temp"
TVOC = "tvoc"
